package eap.freeciv;

import javax.xml.parsers.*;
import org.xml.sax.*;
import org.w3c.dom.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;


import java.io.*;
import java.util.*;

/****************************************************************************
* Generates {@link PacketStructure} objects, by reading XML configuration files.
* <p>
* Right now I'm too lazy to describe the XML
* format tags in detail. Hopefully they are easy to figure out by
* looking at some examples in eap/freeciv/config.
* I created the XML files by
* writing a perl script to parse the freeciv packets.c and packets.h files
* and output a best-guess at the XML. That handled maybe 90% of the
* tedium. I then cleaned things up by hand - mosly for arrays and
* optional fields. So far I have only done this for freeciv 1.13.0, since that is the
* version I have installed on my machine.
* <p>
* There is one configuration file listing the ID and name for each packet.
* Freeciv uses the same format for different types of packets,
* so this list also gives the format name.
* This file must be called "list.xml"
* <p>
* The format for each packet is described in a file with the packet
* format name, with ".xml" appended.  Most fields are simple integers or strings.
* There are special cases like arrays, where a set of fields are read
* multiple times, or bit fields where a single byte encodes up to 8 boolean
* values. If there is enough interest, I will need to document the XML
* carefully - and write a DTD for it.
* <p>
* The idea behind this class is to separate the internal representation
* of the packet structure from the config files. That way the internal
* representation can be omtimized for reading large numbers of the same kind of
* packet. Also, this makes it relatively easy to support other kinds of
* configuration files. I believe this is the only class which requires java 1.4
* because of the XML processing stuff. That could be one reason to replace
* it with something else. Or it could be a good reason to upgrade to the 
* latest runtime environment.
****************************************************************************/
public class PacketBroker {

String config;
ClassLoader loader;
DocumentBuilder builder;

Map by_id;
Map structures;

/****************************************************************************
* Create a new packet broker which will look for the configuration files
* in a default location. Currently the default is "eap/freeciv/config/1.13".
****************************************************************************/
public PacketBroker() throws ParserConfigurationException,
                                          SAXException,
                                          IOException { 
    this("eap/freeciv/config/1.13");
}

/****************************************************************************
* Create a new packet broker which will read the configuration from the given
* directory. This class uses its ClassLoader to find the config files,
* so the directory name should conform to the rules of 
* {@link ClassLoader#getResource(String) }, i.e. a "/" separated path
* relative to the CLASSPATH. Typically the configuration
* directory will be under eap/freeciv/config, but it could be anywhere.
* Note that the configuration files can be in a jar and/or server over 
* a network with an applet's bytecode.
****************************************************************************/
public PacketBroker(String config) throws ParserConfigurationException,
                                          SAXException,
                                          IOException   {
    structures = new HashMap();


    /***********************************************************
    * make sire the configurtation directory ends with a slash *
    ***********************************************************/
    if(! config.endsWith("/") ) config = config+"/";
    this.config = config;

    /*****************************************************************
    * get the class loader so we can use it to find the config files *
    *****************************************************************/
    loader = getClass().getClassLoader();

    /*****************************
    * init the XML parsing stuff *
    *****************************/
    builder = DocumentBuilderFactory.newInstance()
                                    .newDocumentBuilder();

    readPacketList();
                                    
} // end of constructor

/*************************************************************************
* Read the given XML file from the config directory, appending ".xml"
* to its name. 
* @return a "DOM" document describing the XML document.
*************************************************************************/
private Document readXMLFile(String name) throws SAXException, IOException {

    /****************************************************************
    * use the class loader to find the file and open it as a stream *
    ****************************************************************/
    InputStream in = loader.getResourceAsStream(config+name+".xml");
    if(in==null) {
        throw new IOException("Could not find "+name+".xml");
    }

    /**************************************
    * parse the XML and return the result *
    **************************************/
    return builder.parse(in);

} // end of readXMLFile

/*************************************************************************
* read the XML file containing a list of packets
*************************************************************************/
private void readPacketList() throws SAXException, IOException {

    by_id = new HashMap();

    /********************
    * read the XML file *
    ********************/
    Document doc = readXMLFile("list");

    /****************************
    * loop over all the packets *
    ****************************/
    NodeList nodes = doc.getElementsByTagName("PACKET");
    for(int i=0; i<nodes.getLength(); ++i) {
        Element element = (Element)nodes.item(i);
        

        Packet packet = new Packet();

        packet.name = element.getElementsByTagName("NAME")
                            .item(0)
                            .getFirstChild()
                            .getNodeValue();
                            
        packet.id = Integer.parseInt(element.getElementsByTagName("ID")
                                        .item(0)
                                        .getFirstChild()
                                        .getNodeValue()  );

         packet.type = element.getElementsByTagName("STRUCTURE")
                                  .item(0)
                                  .getFirstChild()
                                  .getNodeValue();
                                  
         by_id.put(new Integer(packet.id), packet);


        // System.out.println(packet);
         

      //  getStructure(packet.type);
        

    } // end of loop over packets
        



} // end of constructor

/****************************************************************************
* Return the structure corresponding to a given structure name. Note that this
* is different from the packet name, since freeciv uses the same format for
* different packet types. This method is mostly for internal use.
****************************************************************************/
protected PacketStructure getStructure(String structure_name) throws SAXException, IOException {

    /****************************************
    * see if we already have this structure *
    ****************************************/
    PacketStructure structure = (PacketStructure)structures.get(structure_name);
    if(structure != null) return structure;

    /**************************************************
    * if we get here, we have to set up the structure *
    **************************************************/
    Document doc = readXMLFile(structure_name);
    structure = new PacketStructure();

    /******************
    * loop over nodes *
    ******************/
    for(Node node = doc.getDocumentElement().getFirstChild();
        node != null;
        node = node.getNextSibling() ) {

        if(node.getNodeType() != Node.ELEMENT_NODE) continue;
        Element element = (Element)node;

        structure.addField(parseField(element));

    } // end of loop over elements


    return structure;

} // end of createPacket method

/****************************************************************************
* Returns the structure for a given Packet ID. The ID is the number encoded
* in the third byte of the packet header. Note this method can return the
* same PacketStructure each time it is called with the same ID. In fact that's
* the whole idea of separating the structure from the storage, so that we can
* read the configuration once and cache the structures.
****************************************************************************/
public PacketStructure getStructure(int id) throws IOException {


    Packet packet = (Packet)by_id.get(new Integer(id));
    if(packet==null) throw new IOException("Unknown packet ID "+id);

    try { return getStructure(packet.type); }
    catch(SAXException e) {
        throw new IOException("XML error for structure "+packet.type);
    }

} // end of getStructure method

/****************************************************************************
* Returns the packet name (not the format name) for a given ID.
****************************************************************************/
public String getName(int id) throws IOException {


    Packet packet = (Packet)by_id.get(new Integer(id));
    if(packet==null) throw new IOException("Unknown packet ID "+id);
    
    return packet.name;

} // end of getName method


/****************************************************************************
* Returns the Field object corresponding to a given XML element.
* This method calls one of a number of specialized methods depending on the
* element name.
****************************************************************************/
private Field parseField(Element element) throws IOException {

    /*******************
    * get the tag name *
    *******************/
    String tag = element.getNodeName();
   // System.out.println(tag);

    /***************************************
    * use the tag name to determine how to parse the element *
    *********************************************************/
    if(     tag.equals("FIELD")       ) return parseSimpleField(element);
    else if(tag.equals("BIT_FIELD")   ) return    parseBitField(element);
    else if(tag.equals("ARRAY")       ) return  parseArrayField(element);
    else if(tag.equals("BINARY_FIELD")) return parseBinaryField(element);
    else throw new IOException("Unknown tag "+tag);

} // end of parseField method

/****************************************************************************
* Return a Field object for an XML <FIELD> element.
****************************************************************************/
private SimpleField parseSimpleField(Element element) throws IOException {

    /****************************
    * get the name of the field *
    ****************************/
    String name = element.getElementsByTagName("NAME")
                        .item(0)
                        .getFirstChild()
                        .getNodeValue();

    /********************
    * get the data type *
    ********************/
    String type = element.getElementsByTagName("TYPE")
                        .item(0)
                        .getFirstChild()
                        .getNodeValue();

    /**************************************************
    * determine the size of the field if there is one *
    **************************************************/
    int size=0;
    try {
        size = Integer.parseInt(element.getElementsByTagName("SIZE")
                                    .item(0)
                                    .getFirstChild()
                                    .getNodeValue()   );
    } catch(Exception e) {}

    /**************************
    * is this field optionsl? *
    **************************/
    boolean optional = !element.getAttribute("optional").equals("");
    
    /********************************
    * is this field only read sometimes? *
    *************************************/
    String condition = element.getAttribute("if");
    
    /***************
    * scale factor *
    ***************/
    double scale;
    NodeList scale_nodes =element.getElementsByTagName("SCALE");
    if(scale_nodes.getLength()==0) {
        scale=1.0;
    } else if(scale_nodes.getLength() == 1) {

        String scale_string = scale_nodes.item(0)
                                         .getFirstChild()
                                         .getNodeValue();
                                         
        scale = Double.parseDouble(scale_string);
    } else {
        throw new IOException("Can't have multiple SCALE elements\n");
    }

    /***********************************
    * create a field based on the type *
    ***********************************/
    SimpleField field = null;
    if(     type.equals("string")    ) field = new StringField();
    else if(type.equals("uint")      ) field = new UnsignedIntField(size/8);
    else if(type.equals("sint")      ) field = new   SignedIntField(size/8);
    else if(type.equals("bool")      ) field = new     BooleanField(size/8);
    else if(type.equals("bit_string")) field = new BitStringField();
    else if(type.equals("worklist")  ) field = new WorkListField();
    else if(type.equals("city_map")  ) field = new CityMapField();
    else if(type.equals("tech_list") ) field = new TechListField();
    else {
        throw new IOException("Unknown field type "+type);
    }



    /********************************************************
    * set the name of the field and add it to the structure *
    ********************************************************/
    field.setName(name);
    field.setOptional(optional);
    field.setCondition(condition);
    field.setScale(scale);

    return field;


} // end of parseSimpleField method

/****************************************************************************
* Return a Field object for an XML <BIT_FIELD> element.
****************************************************************************/
private BitField parseBitField(Element element) throws IOException {

    BitField field = new BitField();

    NodeList nodes = element.getElementsByTagName("NAME");
    for(int i=0; i<nodes.getLength(); ++i) {
        Element bit = (Element)nodes.item(i);
        String name = bit.getFirstChild().getNodeValue();
        int index = Integer.parseInt(bit.getAttribute("bit"));

       // System.out.println("    bit="+index+"  name="+name);

        field.setName(name, index);
    }


    return field;

} // end of parseBitField method

/****************************************************************************
* Return a Field object for an XML <ARRAY> element.
****************************************************************************/
private ArrayField parseArrayField(Element array_element) throws IOException {

    String repeat = array_element.getAttribute("repeat");
    ArrayField array = new ArrayField(repeat);


    /******************
    * loop over nodes *
    ******************/
    for(Node node = array_element.getFirstChild();
        node != null;
        node = node.getNextSibling() ) {

        if(node.getNodeType() != Node.ELEMENT_NODE) continue;
        Element element = (Element)node;

        array.addField(parseField(element));

    } // end of loop over elements
    

    return array;

} // end of parseArrayField method

/****************************************************************************
* Return a Field object for an XML <BINARY> element.
****************************************************************************/
public BinaryField parseBinaryField(Element element) throws IOException {

    String name = element.getElementsByTagName("NAME")
                         .item(0)
                         .getFirstChild()
                         .getNodeValue();
                         
    String size = element.getElementsByTagName("SIZE")
                         .item(0)
                         .getFirstChild()
                         .getNodeValue();
                         
    BinaryField field = new BinaryField();
    field.setName(name);
    field.setSize(size);
    
    return field;

} // end of parseBinaryField method

/****************************************************************************
* A simple class for holding the name, ID, and format name for a packet.
****************************************************************************/
private class Packet {

    public String name;
    public int id;
    public String type;

    public String toString() { return id+" "+name+" "+type; }

} // end of Packet inner class


} // end of PacketBroker class
