package eap.freeciv;

import java.io.*;

/*******************************************************************************
* Allows you to read packets from an InputStream.
* All of the {@link Field}
* objects read their data from a PacketSource.
*******************************************************************************/
public class PacketSource {

PacketBroker broker;
InputStream in;
int left;

private static final boolean CHECK_PACKETS=true;


/****************************************************************************
* Creates a new packet source which will read from the given underlying stream
* and use the given packet broker to determine the structure of each
* packet.
****************************************************************************/
public PacketSource(InputStream in, PacketBroker broker) {

    this.in = in;
    left=0;

    this.broker = broker;

} // end of constructor

/****************************************************************************
* Creates a new packet source which will use a default packet broker.
****************************************************************************/
public PacketSource(InputStream in) {

    this.in = in;
    left=0;

    try { this.broker = new PacketBroker(); }
    catch(Exception e) { System.out.println(e); }


} // end of constructor

/***************************************************************************
* Returns the number of unread bytes in the current packet, according to
* the size field in the packet header.
***************************************************************************/
public int remainingBytes() { return left; }

/***************************************************************************
* Reads the next packet from the data stream. This method blocks until
* the next packet can be read. 
* <p>
* Currently, this method checks each packet for validity by saving the
* packet bytes verbatim from the stream, and then writing the packet to
* a byte array and comparing the two. If they are different, it prints
* an error message to stderr and exits. You can turn this off by setting
* CHECK_PACKETS and recompiling.
*
* @return the next packet or null if the underlying stream was at EOF when
*         this method was called.
* @throws EOFException if the underlying stream reached EOF in the middle of 
*         the packet.
* @throws IOException if the data did not conform to the proper structure, 
*         such as if the actual packet size was different from the size in the
*         packet header. I should really create a new exception for Packet 
*         structure problems, but for now it was easier to use IOException
*         for everything.
***************************************************************************/
public Packet readPacket() throws IOException {

    /***********************
    * read the packet size *
    ***********************/
//System.out.println("available="+in.available());
    int hi = in.read();

    int lo = in.read();

    if(hi==-1) return null;
    if(lo==-1) throw new EOFException();

    left = (hi <<8) | lo;
    
    left -= 2;

    /****************
    * read the type *
    ****************/
    int id = in.read();
    if(id==-1) throw new EOFException();
    
    left -= 1;
    
    /*********************************
    * read all the data into a buffer
    * for test purposes
    **********************************/
    InputStream save_in = null;
    byte[] original =null;
    if(CHECK_PACKETS) {
        original = new byte[left];
        if(in.read(original) != left) {
            System.err.println("Couldn't get it all");
            System.exit(1);
        }

        save_in = in;
        in = new ByteArrayInputStream(original);
    }

    
   // System.out.println("reading id="+id+" "+broker.getName(id)+" left="+left);

    /***************************
    * get the packet structure *
    ***************************/
    PacketStructure structure = broker.getStructure(id);
    String name = broker.getName(id);
    
    Packet packet = new Packet(name, id, structure);
    packet.readBody(this);


    if(left != 0 ) throw new IOException(left+" bytes unread in packet");
    
    
    /********************************************************
    * now write the packet to another buffer and compare
    * for text purpoaes
    *******************************************************/
    if(CHECK_PACKETS) {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        packet.write(out);
        byte[] written = out.toByteArray();

        if(original.length != written.length-3) {
            System.err.println("||||||||||||||||||||||||||||||||||||||||||");
            System.err.println(packet);
            packet.dump(System.err);
            System.err.println("written packet has a different length");
            System.err.println("original="+original.length+" written="+(written.length-3));
            System.exit(1);
        }

        if((byte)hi != written[0] || (byte)lo != written[1] ) {
            System.err.println("||||||||||||||||||||||||||||||||||||||||||");
            System.err.println(packet);
            packet.dump(System.err);
            System.err.println("written size is not correct");
            System.err.println("original hi="+hi+" lo="+lo);
            System.err.println("written hi="+written[0]+" lo="+(((int)written[1])&0xff));
            System.exit(1);
        }

        int written_size = (written[0] << 8) | written[1];


        for(int i=0; i<original.length; ++i) {
            if(original[i] != written[i+3] ) {
                System.err.println("||||||||||||||||||||||||||||||||||||||||||");
                System.err.println(packet);
                packet.dump(System.err);
                System.err.println("original and written differ at byte "+i);
                System.err.println("original="+original[i]+" written="+written[i+3]);
                System.exit(1);
            }
        }

        // System.out.println(packet+" is OK");
        in = save_in;
    } // end if we are checking each packet

    return packet;


} // end of readPacket method

/***************************************************************************
* Reads the next byte in the packet. All of the {@link Field} classes use
* this method or {@link #readBytes(byte[])} to read from the underlying stream.
* These methods keep track of the "bytes reamining" accounting.
* This method blocks until the next byte is available.
* @return the next byte in the stream as an unsigned integer.
* @throws IOException if you try to read past the end of the current packet
* @throws EOFException if you try to read when you are at EOF.
***************************************************************************/
public int readByte() throws IOException {

    if(left<=0) throw new IOException("Attempt to read past the end of packet");

    int value = in.read();
    if(value == -1) throw new EOFException();

    --left;

    return value;

} // end of readByte method

/***************************************************************************
* Fill a byte array by reading from the stream. This method blocks until
* all the bytes are read.
* @throws IOException if you try to read past the end of the current packet
* @throws EOFException if you reach EOF.
***************************************************************************/
public void readBytes(byte[] data) throws IOException {

    if(left < data.length) throw new EOFException();
    
    for(int i=0; i<data.length; ++i) {
        data[i] = (byte)readByte();
    }

} // end of readBytes method

} // end of PacketSource class
