package eap.freeciv;

import java.io.*;
import java.util.*;

/***************************************************************************
* Describes the way a packet is encoded in a data stream.
* A structure consistes of an ordered list of {@link Field} objects.
* Generally you get a structure from a {@link PacketBroker}, which in turn reads
* a configuration file. Note that the structure does not
* store any of the packet data. However, the structure knows how to read the
* data from a stream and store it in a {@link Packet}. By separating storage
* and structure, we can create one structure object for each packet type,
* and reuse it for every packet of that type. This avoids having to re-read
* the configuration for every packet.
***************************************************************************/
public class PacketStructure {

List fields;

/***************************************************************************
* Create a new structure with no fields.
***************************************************************************/
public PacketStructure() {

    fields = new ArrayList();


}  // end of constructor

/***************************************************************************
* Append a field to the end of this structure.
***************************************************************************/
public void addField(Field field) {
    fields.add(field);
}

/***************************************************************************
* Read the packet from the given source and store its values in the given
* packet.
***************************************************************************/
public void read(Packet packet, PacketSource source) throws IOException {

    for(Iterator it=fields.iterator(); it.hasNext(); ) {
        Field field = (Field)it.next();

        field.read(packet, source);
    }
} // end of read method

/***************************************************************************
* Write the body of the given packet to the given output stream.
* Note this does not write the packet header. {@link Packet#write(OutputStream)}
* writes the packet header and then calls this method.
***************************************************************************/
public void write(Packet packet, OutputStream out) throws IOException {

    for(Iterator it=fields.iterator(); it.hasNext(); ) {
        Field field = (Field)it.next();
        
 //       System.out.println("writing field");

        field.write(packet, out);
    }
} // end of write method

/***************************************************************************
* Calculate the size of the body of a given packet in bytes. Note that we need 
* both the
* data and the structure to determine the size of a packet, because
* some fields are optional and some have variable size. Note that the packet
* header has three bytes, so you need to add three to this number to get the
* packet size encoded in the header.
***************************************************************************/
public int size(Packet packet) {

    int size=0;
    for(Iterator it=fields.iterator(); it.hasNext(); ) {
        Field field = (Field)it.next();
        
 //       System.out.println("field size = "+field.getSize(packet) );

        size += field.getSize(packet);
    }
    
    return size;

} // end of size method

} // end of PacketStructure class
