package eap.freeciv;

import java.io.*;

/****************************************************************************
* Represents a field with a name and a single value. Note that "single value"
* means that its value is represented by a single object in a packet, so 
* it includes things like a WorkList, where several values are encoded in
* a single object.
* This is an abstract class which handles a number of general properties
* of fields, but does not understand the actual data format.
* <p>
* A field may be optional, meaning it should be written only if it is defined
* in the packet, and it should be read only if there are bytes remaining
* in the packet. 
* <p>
* A field may be present only if a given condition is met. The definition of
* this condition is a bit nebulous - so the best way to understand it would be
* to read the source code for {@link #testCondition(Packet)}. I think
* there is only one field in all of freeciv which uses this.
* <p>
* A field may have a scale factor. Numerical valued fields
* should multiply the raw value by this scale factor.
*
****************************************************************************/
public abstract class SimpleField extends Field {

String name;
boolean optional;
String condition;
double scale;

/****************************************************************************
* Create a new simple field. Initializes the scale to 1.0, the condition to 
* none, and set the field as required.
****************************************************************************/
protected SimpleField() {

    scale=1.0;
    optional=false;
    condition=null;

} // end of constructor

/****************************************************************************
* Set the name of this field. This names is used as the keyword for accessing
* the data in a packet.
****************************************************************************/
public void setName(String name) { this.name = name; }

/****************************************************************************
* Returns the name of this field.
****************************************************************************/
public String getname() { return name; }

/****************************************************************************
* Adds the given index in brackets to the end of the name of this field.
* This is used by arrays.
* @see Field#addIndex(int)
****************************************************************************/
public void addIndex(int index) {

    name = name + "["+index+"]";
}


/****************************************************************************
*
****************************************************************************/
public void removeIndex() {

    int bracket = name.lastIndexOf('[');
    if(bracket==-1) return;
    name = name.substring(0,bracket);
}

/****************************************************************************
* Returns true if the packet has a value with the same name as this field.
****************************************************************************/
public boolean isSet(Packet packet) {

    return packet.hasValue(name);
}

/****************************************************************************
* Set whether this field is optional
****************************************************************************/
public void setOptional(boolean optional) { this.optional = optional; }

/****************************************************************************
* Indicates whether this field is optional.
****************************************************************************/
public boolean isOptional() { return optional; }

/****************************************************************************
* Set a string describing a condition for whether this field should appear
* in the packet.
* @see #testCondition(Packet)
****************************************************************************/
public void setCondition(String condition) { this.condition = condition; }

/****************************************************************************
* Set the scale factor for this packet. Numerical-valued subclasses should
* multiply raw values by this scale factor.
****************************************************************************/
public void setScale(double scale) { this.scale = scale; }

/****************************************************************************
* Read the data in this field from the packet. The {@link #read} method
* determines whether this field should appear in the packet and then calls
* this method to read the actual data.
****************************************************************************/
protected abstract void readValue(Packet packet, PacketSource source) throws IOException;

/****************************************************************************
* Write the data in this field. The {@link #write} method
* determines whether this field should appear in the packet and then calls
* this method to write the actual data.
****************************************************************************/
protected abstract void writeValue(Object value, OutputStream out) throws IOException;

/****************************************************************************
* Returns the size of the data in this field. The {@link #getSize(Packet)}
* method determines whether the packet contains this field and then
* calls this method to return its size.
****************************************************************************/
public abstract int getValueSize(Packet packet);

/****************************************************************************
* Read the value from the packet if it is present. Subclasses should override
* {@link #readValue(Packet, PacketSource)} to specify how the data are
* encoded.
****************************************************************************/
public final void read(Packet packet, PacketSource source) throws IOException {

  //  System.out.println("reading "+name+" left="+source.remainingBytes() );

    if(optional && source.remainingBytes() == 0 ) return;

    if(testCondition(packet)) readValue(packet, source);

} // end of read method

/****************************************************************************
* Write this field if it is present in the packet. Subclasses should override
* {@link #writeValue(Object, OutputStream)} to specify how the data are
* encoded.
****************************************************************************/
public final void write(Packet packet, OutputStream out) throws IOException {

    Object value = packet.getValue(name);
    if(optional && value == null) return;

    if(testCondition(packet) ) {
        if(value == null) throw new IOException(name+" is not set in "+packet);
        writeValue(value, out);
    }

} // end of write method

/****************************************************************************
* Evaluate a test condition to see if a packets contains this field.
* This is rarely used.
****************************************************************************/
protected boolean testCondition(Packet packet) throws IOException {

    if(condition==null)      return true;
    if(condition.equals("")) return true;

    if(condition.indexOf("(")!= -1) {
        /***************************
        * have to apply a function *
        ***************************/
        if(condition.startsWith("bit(") ) {
            /****************
            * extract a bit *
            ****************/
            int left_bracket  = condition.indexOf("(");
            int comma         = condition.indexOf(",");
            int right_bracket = condition.indexOf(")");
            
            String name = condition.substring(left_bracket+1, comma).trim();
            int bit = Integer.parseInt(condition.substring(comma+1, right_bracket).trim());
            
            long value = ((Number)packet.getValue(name)).longValue();
            return (value & (1<<bit)) != 0;
        } else {
            throw new IOException("Unknown function in condition "+condition);
        }
    } else {
        /*************************
        * just a plain condition *
        *************************/
        Object value = packet.getValue(condition);
        if(value == null) throw new IOException("invalid condition "+condition);
        if(value instanceof Boolean) return ((Boolean)value).booleanValue();
        else if(value instanceof Number ) {
            return ((Number)value).longValue() != 0l;
        } else {
            throw new IOException("Can't convert "+condition+"="+value+
                                  " to a boolean");
        }
    }

} // end of testCondition method

/****************************************************************************
* Returns the size of this field. If the field is not included in the packet,
* then this method returns zero. Otherwise it calls {@link #getValueSize(Packet)}
* to determine the size.
****************************************************************************/
public final int getSize(Packet packet) {

    Object value = packet.getValue(name);
    if(optional && value == null) return 0;

    try { if(!testCondition(packet) ) return 0; }
    catch(IOException e) { 
        System.err.println(e);
        return 0; 
    }
    
    return getValueSize(packet);
    
} // end of getSize method

} // end of SimpleField class
