package eap.freeciv;

import java.io.*;
import java.util.*;

/****************************************************************************
* Represents a set of repeated fields in a packet.
* An array is composed of zero or more elements. Each element is composed
* of one or more fields. Each element has the same fields. These fields can be
* of any type, including other arrays - though
* currently the freeciv packets don't do that.
* <p>
* When referencing the data in a packet, this class indicates the element by
* calling {@link Field#addIndex(int)}. That means that in general, you indicate
* the element you want by appending "[index]" to the keyword for a datum in 
* a packet.
* <p>
* There are three ways to specify the number of elements in an array.
* <ol>
* <li> The array can have a fixed number of elements
* <li> The number of elements can be specified by a field in the packet
* <li> The array continues to the end of the packet. If you are not reading
*      a packet, then the array continues until you reach an element for
*      which at least one field is unset.
* </ol>
****************************************************************************/
public class ArrayField extends Field {

List fields;
int repeat;
String repeat_field;

/****************************************************************************
* Create a new array with the given repeat count. The repeat count is
* interpreted as by {@link #setRepeat(String) }.
****************************************************************************/
public ArrayField(String repeat) {
    fields = new ArrayList();
    setRepeat(repeat);
}


/****************************************************************************
* Add an index to this file. This method just calls addIndex for all of the
* fields contained in the array. This method would only be used if an array
* contained another array, which currently doesn't happen in freeciv.
****************************************************************************/
public void addIndex(int index) {

    for(Iterator it=fields.iterator(); it.hasNext(); ) {
        Field field = (Field)it.next();

        field.addIndex(index);
    }
} // end of addIndex method

/****************************************************************************
* calls removeIndex for all of the fields in this array.
* @see #addIndex(int)
****************************************************************************/
public void removeIndex() {

    for(Iterator it=fields.iterator(); it.hasNext(); ) {
        Field field = (Field)it.next();

        field.removeIndex();
    }
} // end of removeIndex method

/****************************************************************************
* Returns true if all the fields for all the elements in the array are
* set in the given packet. Note that if this array "fills" to the end of
* the packet, this method will always return true.
****************************************************************************/
public boolean isSet(Packet packet) {

    for(ElementIterator elem= new OutputElementIterator(packet);
        elem.hasElement(); elem.increment() ) {

        for(Iterator it=fields.iterator(); it.hasNext(); ) {
            Field field = (Field)it.next();

            field.addIndex(elem.getIndex());
            if(! field.isSet(packet) ) return false;
            field.removeIndex();
        }
    } // end of loop over elements
    
    return true;

} // end of isSet method

/****************************************************************************
* Sets the number of elements according to a given string. If the string can be
* parsed  by {@link Integer#parseInt(String) }, then it is interpreted as
* a fixed number of elements. If the string is null or empty, then the array
* continues to the end of the packet. Otherwise the string is
* assumed to be the name of a field in the packet. The value of this field
* will be interpreted as an integer indicating the repeat count.
****************************************************************************/
public void setRepeat(String repeat) {

    this.repeat = -1;
    repeat_field = null;

    /***********************************************
    * no value means read to the end of the packet *
    ***********************************************/
    if(repeat == null || repeat.equals("")) return;

    /**********************************
    * either a number of a field name *
    **********************************/
    try { this.repeat = Integer.parseInt(repeat); }
    catch(NumberFormatException e) {
        repeat_field = repeat;
    }
} // end of setRepeat method

/****************************************************************************
* Append a field to the end of the list of fields in this array.
****************************************************************************/
public void addField(Field field) {

    fields.add(field);
}

/****************************************************************************
*
****************************************************************************/
public void read(Packet packet, PacketSource source) throws IOException {

    for(ElementIterator elem= new InputElementIterator(packet, source);
        elem.hasElement(); elem.increment() ) {

        /**************************************
        * loop over the fields in the element *
        **************************************/
        for(Iterator it=fields.iterator(); it.hasNext(); ) {
            Field field = (Field)it.next();

            field.addIndex(elem.getIndex());
            field.read(packet, source);
            field.removeIndex();

        }  // end of loop over fields
    } // end of loop over elements

} // end of read method


/****************************************************************************
*
****************************************************************************/
public void write(Packet packet, OutputStream out) throws IOException {

    for(ElementIterator elem= new OutputElementIterator(packet);
        elem.hasElement(); elem.increment() ) {

        for(Iterator it=fields.iterator(); it.hasNext(); ) {
            Field field = (Field)it.next();

            field.addIndex(elem.getIndex());
            field.write(packet, out);
            field.removeIndex();
        }
    } // end of loop over elements



} // end of write method


/****************************************************************************
* Sum the sizes of all the elements in this array. Each element has a size
* equal to the sum of the sizes of its fields.
****************************************************************************/
public int getSize(Packet packet) {

    int size=0;
    for(ElementIterator elem= new OutputElementIterator(packet);
        elem.hasElement(); elem.increment() ) {

        for(Iterator it=fields.iterator(); it.hasNext(); ) {
            Field field = (Field)it.next();

            field.addIndex(elem.getIndex());
            size += field.getSize(packet);
            field.removeIndex();
        }
    }

    return size;

} // end of getSize method

/*************************************************************************
* This is a convenience class for iterating over the elements in an array
* This is necessary, because it is non-trivial to determine the number of
* elements in an array.
*************************************************************************/
private abstract class ElementIterator {

    int index;
    int nelements;
    boolean nelements_known;

    public ElementIterator(Packet packet) {

        index=0;
        /*******************************************
        * determine the number of elements to read *
        *******************************************/
        nelements_known = repeat>=0 || repeat_field!=null;

        if(repeat_field==null) nelements = repeat;
        else  nelements = ((Number)packet.getValue(repeat_field)).intValue();

    } // end of constructor

    public int getIndex() { return index; }
    public void increment() { ++index; }
    public boolean hasElement() {
        if(nelements_known) return index < nelements;
        else                return checkForElement();
    }

    protected abstract boolean checkForElement();

} // end of ElementIterator inner class

/*************************************************************************
* Use this kind of iterator when reading from a packet
*************************************************************************/
private class InputElementIterator extends ElementIterator {

    private PacketSource source;

    public InputElementIterator(Packet packet, PacketSource source) {
        super(packet);
        this.source = source;
    }

protected boolean checkForElement() { return source.remainingBytes()>0; }

} // end of InputElementIterator inner class

/*************************************************************************
* Use this kind of iterator when you are not reading.
*************************************************************************/
private class OutputElementIterator extends ElementIterator {

    private Packet packet;

    public OutputElementIterator(Packet packet) {
        super(packet);
        this.packet = packet;
    }

    protected boolean checkForElement() {

        for(Iterator it=fields.iterator(); it.hasNext(); ) {
            Field field = (Field)it.next();

            field.addIndex(index);
            boolean set = field.isSet(packet);
            field.removeIndex();

            if(!set) return false;

        }

        return true;
    } // end of checkForElement method

} // end of InputElementIterator inner class

} // end of ArrayField class
