package eap.freeciv;

import java.io.*;
import java.net.*;
import java.util.*;

/********************************************************************************
* Represents a socket connection through which freeciv packets will flow.
* The main purpose of this class is to run a thread which listens for packets 
* from the socket and passes them to registered {@link PacketListener}
* objects.
* Such a connection may be on either the client or server side.
* <p>
* The typical way to use this class is to create a {@link Socket}, use it
* to create a PacketConnection, register listeners with 
* {@link #addPacketListener(PacketListener)}, and then call {@link #beginListening()}
* to start the packet listening thread.
* Usually, a {@link Client} or {@link Server} will handle much of this for you.
********************************************************************************/
public class PacketConnection extends Thread {

PacketSource source;
OutputStream out;
Set listeners;

/********************************************************************************
* Create a new connection associated with the given socket.
********************************************************************************/
public PacketConnection(Socket socket) throws IOException {

    source = new PacketSource(socket.getInputStream());
    out = socket.getOutputStream();

    listeners = new HashSet();

} // end of constructor

/*****************************************************************************
* Send a packet through the socket.
*****************************************************************************/
public void sendPacket(Packet packet) throws IOException {

    packet.write(out);

} // end of sendPacket method

/*****************************************************************************
* Register a listener to receive the packets that this class reads from the
* socket.
*****************************************************************************/
public void addPacketListener(PacketListener l) { listeners.add(l); }

/*******************************************************************************
* Starts the listening thread. Note that you should register all the packet
* listeners before calling this method. Otherwise, the connection might
* read some packets and discard them before you begin listening.
* <p>
* Here's a thought - maybe the thread should wait if there are no listeners
* registered. The only downside is that the a connection could never just 
* discard packets - but I can't think of a reason you would need it to, and
* even if you did you could just create a "/dev/null" packet listener.
*******************************************************************************/
public void beginListening() { start(); }

/*******************************************************************************
* This method is required by the Thread class. It implements the listening
* thread.
*******************************************************************************/
public void run() {

    try {
        while(true) {

            /*******************************************************
            * read the next packet and check for a comfortable EOF *
            *******************************************************/
            Packet packet = source.readPacket();
            if(packet==null) return;
            
            /***************************************
            * send the packet to all the listeners *
            ***************************************/
            for(Iterator it=listeners.iterator(); it.hasNext(); ) {
                PacketListener l = (PacketListener)it.next();
                
                l.receivePacket(packet);
            }
        } // end of loop over packets
    } catch(IOException e) {
        System.err.println("Connection caught exception while listening");
        System.err.println(e);
        e.printStackTrace();
        System.exit(1);
    }

} // end of run method

} // end of PacketConnection class
